"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.Form = void 0;
const constants_1 = require("./constants");
const utils_1 = require("./utils");
class Form {
    constructor(parent, options) {
        this.options = {
            id: '',
            saveProgress: false,
            useFormData: false,
            schema: [],
            action: null,
            method: null,
            className: constants_1.FORM_CLASS_DEFAULT,
        };
        this._parent = null;
        this._formElement = null;
        this._id = '';
        this._isValid = null;
        this._errors = [];
        this._fields = {};
        this._groups = {};
        this._buttons = {};
        this._data = {};
        this._saveProgress = false;
        this._licenseState = constants_1.LICENSE_STATE.INVALID;
        this._schema = [];
        this._dataPrefixMap = {};
        this._triggerEvents = true;
        this.initializeOptions(options);
        this.setParentElement(parent);
        this.processLicense();
        this.onGui();
        this.initForm();
    }
    setParentElement(parent) {
        let parentElement;
        if (typeof parent === 'string') {
            parentElement = document.getElementById(parent);
        }
        else {
            parentElement = parent;
        }
        this._parent = parentElement;
        if (!this._parent)
            throw new Error('Parent does not exist in the document');
    }
    initializeOptions(options) {
        this.options = Object.assign({}, this.options, options);
        this._id = options.id;
        this._schema = options.schema;
        this._saveProgress = options.saveProgress;
        if (this._saveProgress) {
            (0, utils_1.useLicensedFetures)();
        }
    }
    initForm() {
        if (this._formElement)
            this.buildSchema(this._schema, this._formElement);
        if (this._formElement && this._parent)
            (0, utils_1.mountElement)(this._formElement, this._parent);
    }
    processLicense() {
        if (!(0, utils_1.usesLicensedFetures)())
            return;
        if (this.options.licenseKey)
            (0, utils_1.setLicenseKey)(this.options.licenseKey);
        this._licenseState = (0, utils_1.processLicenseKey)();
        if (!this.hasValidLicense()) {
            (0, utils_1.handleInvalidLicenseLog)(this._licenseState);
        }
    }
    buildSchema(schema, parent, groupId = null, listId = null, key = null) {
        schema.forEach((options) => {
            this.buildElement(options, parent, groupId, listId, key);
        });
    }
    isValid() {
        return this._isValid;
    }
    createWrapper(parent) {
        const wrapper = document.createElement(constants_1.DIV_ELEMENT);
        (0, utils_1.mountElement)(wrapper, parent);
        return wrapper;
    }
    mapFieldToDataPrefix(options, id, key = null) {
        let newId = options.id;
        if (key) {
            newId = (0, utils_1.transformFieldName)(id, key, options.id);
        }
        this._dataPrefixMap[newId] = {
            id,
            dataKey: options.id,
            key,
        };
        if (key) {
            options.id = newId;
            options.name = newId;
        }
    }
    buildElement(options, parent, groupId, listId, key) {
        const duplicatedOptions = Object.assign({}, options);
        const Constructor = constants_1.elementConstructors[options.type];
        if (!Constructor)
            throw new Error(`Unknown type: ${options.type}`);
        const wrapper = this.createWrapper(parent);
        const formElementType = (0, utils_1.getFormElementType)(options.type);
        if (listId && key && formElementType === constants_1.constructorTypes.field) {
            this.mapFieldToDataPrefix(duplicatedOptions, listId, key);
        }
        if (groupId && formElementType === constants_1.constructorTypes.field) {
            this.mapFieldToDataPrefix(duplicatedOptions, groupId, null);
        }
        const Constructed = new Constructor(wrapper, this, duplicatedOptions);
        switch (formElementType) {
            case constants_1.constructorTypes.button:
                if (listId && key) {
                    this.assignToListField(listId, key, Constructed, formElementType, options.id);
                }
                else
                    this._buttons[options.id] = Constructed;
                break;
            case constants_1.constructorTypes.group:
                if (listId && key) {
                    this.assignToListField(listId, key, Constructed, formElementType, options.id);
                }
                else
                    this._groups[options.id] = Constructed;
                if (options.schema && Constructed.getSchemaContainer) {
                    const newParent = Constructed.getSchemaContainer();
                    if (newParent) {
                        this.buildSchema(options.schema, newParent, options.prefixSchema ? Constructed.getId() : groupId, listId, key);
                    }
                }
                break;
            case constants_1.constructorTypes.field:
                if (listId && key) {
                    this.assignToListField(listId, key, Constructed, formElementType, options.id);
                }
                else {
                    this._fields[options.id] = Constructed;
                }
                break;
        }
    }
    assignToListField(listId, key, constructed, formElementType, id) {
        const list = this._fields[listId];
        if (!list || !list.assignButton || !list.assignGroup || !list.assignField)
            return;
        switch (formElementType) {
            case constants_1.constructorTypes.button:
                list.assignButton(id, key, constructed);
                break;
            case constants_1.constructorTypes.group:
                list.assignGroup(id, key, constructed);
                break;
            case constants_1.constructorTypes.field:
                list.assignField(id, key, constructed);
                break;
        }
    }
    update() {
        return __awaiter(this, void 0, void 0, function* () {
            if (this._groups) {
                Object.keys(this._groups).forEach((key) => {
                    this._groups[key].update();
                });
            }
            if (this._fields) {
                Object.keys(this._fields).forEach((key) => {
                    this._fields[key].update();
                });
            }
            if (this._buttons) {
                Object.keys(this._buttons).forEach((key) => {
                    this._buttons[key].update();
                });
            }
        });
    }
    getField(id) {
        return this._fields[id];
    }
    getFormElement() {
        return this._formElement;
    }
    dispatchEvent(event, data = null) {
        var _a;
        if (!this._triggerEvents)
            return;
        const dispatched = data !== null
            ? new CustomEvent(event, {
                detail: data,
            })
            : new CustomEvent(event);
        (_a = this._formElement) === null || _a === void 0 ? void 0 : _a.dispatchEvent(dispatched);
    }
    setData(id, value) {
        if (Object.prototype.hasOwnProperty.call(this._dataPrefixMap, id)) {
            this.setDataFromMap(id, value);
        }
        else {
            this.setSimpleData(id, value);
        }
        this.update();
        this.dispatchEvent("dataUpdated", this._data);
    }
    removeData(id) {
        if (Object.prototype.hasOwnProperty.call(this._dataPrefixMap, id)) {
            this.removeMappedData(id);
        }
        else {
            delete this._data[id];
        }
        this.dispatchEvent("dataUpdated", this._data);
    }
    removeMappedData(id) {
        const fieldMapping = this._dataPrefixMap[id];
        const { dataKey, key, id: prefixDataId } = fieldMapping;
        if (!dataKey || !prefixDataId)
            return;
        if (key) {
            const list = this.getField(prefixDataId);
            if (!list || !list.getKeyIndex)
                return;
            const prefixId = list.getId();
            const keyIndex = list.getKeyIndex(key);
            delete this._data[prefixId][keyIndex][dataKey];
            if (typeof this._data[prefixId][keyIndex] === 'object' &&
                Object.keys(this._data[prefixId][keyIndex]).length === 0) {
                delete this._data[prefixId][keyIndex];
            }
        }
        else {
            const group = this.getGroup(prefixDataId);
            if (!group)
                return;
            const prefixId = group.getId();
            delete this._data[prefixId][dataKey];
            if (typeof this._data[prefixId] === 'object' && Object.keys(this._data[prefixId]).length === 0) {
                delete this._data[prefixId];
            }
        }
    }
    setDataFromMap(id, value) {
        const fieldMapping = this._dataPrefixMap[id];
        const { dataKey, key, id: prefixDataId } = fieldMapping;
        if (!dataKey || !prefixDataId)
            return;
        if (key) {
            const list = this.getField(prefixDataId);
            if (!list || !list.getKeyIndex)
                return;
            const prefixId = list.getId();
            const keyIndex = list.getKeyIndex(key);
            this.ensureDataStructureExists(prefixId, keyIndex);
            this._data[prefixId][keyIndex][dataKey] = value;
        }
        else {
            const group = this.getGroup(prefixDataId);
            if (!group)
                return;
            const prefixId = group.getId();
            this.ensureDataStructureExists(prefixId);
            this._data[prefixId][dataKey] = value;
        }
    }
    setSimpleData(id, value) {
        this._data[id] = value;
    }
    ensureDataStructureExists(listId, keyIndex = null) {
        if (keyIndex === null && !this._data[listId]) {
            this._data[listId] = {};
            return;
        }
        else if (keyIndex || keyIndex === 0) {
            if (!this._data[listId])
                this._data[listId] = [];
            if (!this._data[listId][keyIndex])
                this._data[listId][keyIndex] = {};
        }
    }
    removeListData(listId, index) {
        if (this._data[listId] && this._data[listId][index]) {
            this._data[listId].splice(index, 1);
        }
    }
    getId() {
        return this._id;
    }
    getData() {
        return this._data;
    }
    getButtons() {
        return this._buttons;
    }
    getButton(id) {
        return this._buttons[id];
    }
    getGroups() {
        return this._groups;
    }
    getGroup(id) {
        return this._groups[id];
    }
    savesProgress() {
        return this._saveProgress;
    }
    getErrors() {
        return this._errors;
    }
    hasValidLicense() {
        return this._licenseState === constants_1.LICENSE_STATE.VALID;
    }
    updateError(id, isValid) {
        if (this._errors.includes(id) && isValid) {
            this._errors = this._errors.filter((field) => {
                return field !== id;
            });
        }
        else if (!this._errors.includes(id) && !isValid) {
            this._errors.push(id);
        }
        this._isValid = this._errors.length === 0;
    }
    reset(event) {
        this._triggerEvents = false;
        if (event)
            event.preventDefault();
        if (this._groups) {
            Object.keys(this._groups).forEach((key) => {
                const group = this._groups[key];
                if (group.reset && typeof group.reset === 'function')
                    group.reset();
            });
        }
        if (this._fields) {
            Object.keys(this._fields).forEach((key) => {
                const field = this._fields[key];
                if (field.reset && typeof field.reset === 'function')
                    field.reset();
            });
        }
        if (this._buttons) {
            Object.keys(this._buttons).forEach((key) => {
                const button = this._buttons[key];
                if (button.reset && typeof button.reset === 'function')
                    button.reset();
            });
        }
        this._isValid = null;
        this._triggerEvents = true;
        this.dispatchEvent("resetted");
    }
    validate() {
        if (this._fields) {
            Object.keys(this._fields).forEach((key) => {
                const field = this._fields[key];
                if (field.validate && typeof field.validate === 'function')
                    field.validate();
            });
        }
        this._isValid = this._errors.length === 0;
    }
    save() {
        if (!this._saveProgress || !this.hasValidLicense())
            return;
        if (this._fields) {
            Object.keys(this._fields).forEach((key) => {
                const field = this._fields[key];
                if (field.save && typeof field.save === 'function')
                    field.save();
            });
        }
        if (this._groups) {
            Object.keys(this._groups).forEach((key) => {
                const group = this._groups[key];
                if (group.save && typeof group.save === 'function')
                    group.save();
            });
        }
    }
    load() {
        if (!this._saveProgress || !this.hasValidLicense())
            return;
        if (this._fields) {
            Object.keys(this._fields).forEach((key) => {
                const field = this._fields[key];
                if (field.load && typeof field.load === 'function')
                    field.load();
            });
        }
        if (this._groups) {
            Object.keys(this._groups).forEach((key) => {
                const group = this._groups[key];
                if (group.load && typeof group.load === 'function')
                    group.load();
            });
        }
    }
    onGui() {
        this._formElement = document.createElement(constants_1.FORM_ELEMENT);
        this._formElement.setAttribute(constants_1.ID_ATTRIBUTE, this._id);
        this._formElement.className = this.options.className;
        if (this.options.action)
            this._formElement.setAttribute(constants_1.ACTION_ATTRIBUTE, this.options.action);
        if (this.options.method)
            this._formElement.setAttribute(constants_1.METHOD_ATTRIBUTE, this.options.method);
        this._formElement.addEventListener(constants_1.SUBMIT_ATTRIBUTE, (event) => this.submit(event));
        this._formElement.addEventListener(constants_1.RESET_ATTRIBUTE, this.reset);
        if ((0, utils_1.usesLicensedFetures)() && !this.hasValidLicense())
            this.createInvalidElement();
    }
    createInvalidElement() {
        const invalid = document.createElement(constants_1.LINK_ELEMENT);
        invalid.setAttribute(constants_1.HREF_ATTRIBUTE, constants_1.PACKAGE_LICENSE_URL);
        invalid.className = constants_1.licensePlateClass;
        invalid.style.cssText = constants_1.licensePlateStyle;
        const text = document.createElement(constants_1.SPAN_ELEMENT);
        text.innerText = (0, utils_1.getLicenseText)(this._licenseState);
        (0, utils_1.mountElement)(text, invalid);
        if (this._parent)
            (0, utils_1.mountElement)(invalid, this._parent);
    }
    submit(event) {
        if (!this.options.action && event)
            event.preventDefault();
        this.validate();
        if (!this.isValid()) {
            this.dispatchEvent("validationFailedOnSubmit");
            return;
        }
        let data = this.getData();
        if (this.options.useFormData) {
            data = (0, utils_1.objectToFormData)(this.getData());
        }
        if (this.options.submit)
            this.options.submit(data);
        this.dispatchEvent("submitted", data);
    }
    destroy() {
        if (this._parent)
            this._parent.remove();
    }
    on(event, listener, options) {
        var _a;
        (_a = this._formElement) === null || _a === void 0 ? void 0 : _a.addEventListener(event, listener, options);
    }
    off(event, listener, options) {
        var _a;
        (_a = this._formElement) === null || _a === void 0 ? void 0 : _a.removeEventListener(event, listener, options);
    }
}
exports.Form = Form;
